/* Needed for return_to_launcher() */
#include <esp_system.h>
#include <soc/rtc.h>
#include <soc/rtc_cntl_reg.h>

/* Needed for RP2040 interfacing */
#include <Wire.h>

/* Needed for *_display() */
#include <Adafruit_ILI9341.h> /* Adafruit ILI9341 library */

/* Needed for *_leds() */
#include <FastLED.h> /* library to interface with SK9822 (and other) leds */

/* Workaround to skip esp_ota_get_running_partition() */
extern "C" bool verifyRollbackLater() {
  return true;
}

/* See ESP PIN numbers here:
 * https://github.com/badgeteam/esp32-component-mch2022-bsp/
 *   blob/2da8dbd50679512058820b40bfd6d7a5c529b28c/include/mch2022_badge.h */
#define GPIO_I2C_SDA 22
#define GPIO_I2C_SCL 21
#define GPIO_INT_RP2040 34 /* active LOW */

#define GPIO_SPI_CS_LCD 32
#define GPIO_SPI_DC_LCD 33
#define GPIO_LCD_RESET 25
#define GPIO_LCD_MODE 26

#define GPIO_LED_DATA 5
#define GPIO_SD_PWR 19
#define GPIO_LED_ENABLE GPIO_SD_PWR /* enables power to both SD and LEDs */
#define NUM_LEDS 5

/* I2C addresses */
#define RP2040_ADDR 0x17 /* RP2040 co-processor */

/* See RP register addresses here:
 *   https://github.com/badgeteam/esp32-component-mch2022-rp2040/
 *     blob/45fce00c790a2918bf506dd8af0c0b68db05d031/include/rp2040.h
 * Feed through:
 *   sed -e '1,/RP2040_REG_FW_VER/d;/}/,$d;s/,//' | awk '{printf "#define %s 0x%02x\n",$1,NR}'
 */
#define RP2040_REG_LCD_BACKLIGHT 0x04
#define RP2040_REG_INPUT1 0x06
#define RP2040_REG_ADC_VALUE_VBAT_LO 0x0d
#define RP2040_REG_CHARGING_STATE 0x15
#define RP2040_REG_IR_ADDRESS_LO 0x60

/* Buttons, accessible through the RP2040 through wire I2C */
#define BTN_HOME 0x0001   /* home */
#define BTN_MENU 0x0002   /* menu */
#define BTN_START 0x0004  /* start */
#define BTN_A 0x0008      /* accept */
#define BTN_B 0x0010      /* back */
#define BTN_RES0 0x0020   /* (RP2040_INPUT_FPGA_CDONE) */
#define BTN_RES1 0x0040   /* (RP2040_INPUT_BATTERY_CHARGING) */
#define BTN_SELECT 0x0080 /* select */
#define BTN_LEFT 0x0100   /* joy left */
#define BTN_JOY 0x0200    /* joy press */
#define BTN_DOWN 0x0400   /* joy down */
#define BTN_UP 0x0800     /* joy up */
#define BTN_RIGHT 0x1000  /* joy right */

#define BTN2STATE(button) (button & 0xffff)
#define BTN2BTN(button) ((button >> 16) & 0xffff)
#define BUTTON_PRESSED(button, compare) \
  ((BTN2BTN(button) & compare) && (BTN2STATE(button) & compare) == compare)
#define BUTTON_RELEASED(button, compare) \
  ((BTN2BTN(button) & compare) && (BTN2STATE(button) & compare) == 0)

void init_inputs();
uint32_t read_inputs();
void init_display();
void update_display();
void init_leds_after_display();
void change_leds();
void get_battery_state(bool& charging, float& voltage);
void set_backlight(uint8_t brightness);
void send_ir_nec(uint16_t address, uint8_t command);
void return_to_launcher();

Adafruit_ILI9341 tft = Adafruit_ILI9341(GPIO_SPI_CS_LCD, GPIO_SPI_DC_LCD, GPIO_LCD_RESET);
GFXcanvas16* canvas; /* uses malloc(); cannot be initialized statically */

/* LED configuration of the kite on the top left:
 *   _2_                       _0_
 *  / | \                     / | \
 * 0--1--3  == reordered =>  1--2--3
 *  \ | /                     \ | /
 *   \|/                       \|/
 *    4                         4 */
CHSV led_color[NUM_LEDS];
CRGB leds[NUM_LEDS];
const int led_reorder[NUM_LEDS] = {2, 0, 1, 3, 4}; /* left-to-right */

void setup() {
  Serial.begin(115200);
  Serial.println("hello MCH2022");

  init_inputs();
  init_display();
  init_leds_after_display();

  set_backlight(255); // full on

  /* Initialize the 5 LED colors to a different offset. */
  for (int i = 0; i < NUM_LEDS; ++i) {
    led_color[led_reorder[NUM_LEDS - i - 1]].hue = ((i * 8) % 256);
    led_color[i].saturation = 255;
    led_color[i].value = 255;
  }
}

void loop() {
  if (!digitalRead(GPIO_INT_RP2040)) {
    uint32_t input = read_inputs();
    if (BUTTON_RELEASED(input, BTN_HOME)) {
      return_to_launcher();
    /**
     * See these LG codes:
     * https://github.com/Lucaslhm/Flipper-IRDB/blob/48f26b13715547d579a4f17a3fc0983a11d36776/TVs/LG/LG.ir
     * > name: Power
     * > type: parsed
     * > protocol: NEC
     * > address: 04 00 00 00
     * > command: 08 00 00 00
     */
    } else if (BUTTON_PRESSED(input, BTN_START)) {
      send_ir_nec(0x04, 0x08); // LG and others: power (on/off)
    } else if (BUTTON_PRESSED(input, BTN_LEFT)) {
      send_ir_nec(0x04, 0x01); // LG and others: chan-down
    } else if (BUTTON_PRESSED(input, BTN_RIGHT)) {
      send_ir_nec(0x04, 0x00); // LG and others: chan-up
    } else if (BUTTON_PRESSED(input, BTN_UP)) {
      send_ir_nec(0x04, 0x02); // LG and others: vol-up
    } else if (BUTTON_PRESSED(input, BTN_DOWN)) {
      send_ir_nec(0x04, 0x03); // LG and others: vol-down
    } else if (BUTTON_PRESSED(input, BTN_JOY)) {
      send_ir_nec(0x04, 0x09); // LG and others: mute (on/off)
    } else {
      Serial.printf("input: 0x%08x // 0x<button><state>\r\n", input);
    }
  }

  change_leds();

  //unsigned long t0 = millis();
  update_display(); // takes about 60ms (= 17Hz)
  //Serial.printf("updating display took %lu ms\r\n", millis() - t0);
}

void init_inputs() {
  Wire.begin(GPIO_I2C_SDA, GPIO_I2C_SCL);
  pinMode(GPIO_INT_RP2040, INPUT);
  read_inputs();
}

uint32_t read_inputs() {
  Wire.beginTransmission(RP2040_ADDR);
  Wire.write(RP2040_REG_INPUT1);
  Wire.endTransmission();
  Wire.requestFrom(RP2040_ADDR, 4);
  /* state  = 0x20 + sum of all pressed buttons */
  uint16_t state = Wire.read() | (Wire.read() << 8);
  /* which_button = 1, 2, 4, 8, ... */
  uint16_t which_button = Wire.read() | (Wire.read() << 8);
  /* NOTE: which_button may hold more than one if the presses/releases
   * are registered "simultaneously" */
  return (which_button << 16 | state);
}

void init_display() {
  pinMode(GPIO_LCD_MODE, OUTPUT);
  digitalWrite(GPIO_LCD_MODE, LOW); /* set "ESP32" output instead of "FPGA" */

  tft.begin();
  tft.fillScreen(ILI9341_BLACK);
  tft.setRotation(1);

  /* Using a buffer is slightly slower than writing directly. But it's
   * significantly easier to avoid flicker. */
  canvas = new GFXcanvas16(ILI9341_TFTHEIGHT, ILI9341_TFTWIDTH); /* 16-bit, 320x240 */
  canvas->setRotation(0);
}

void update_display() {
  static int x = 0;
  static int y = 0;

  bool charging;
  float voltage;
  get_battery_state(charging, voltage);

  canvas->fillScreen(ILI9341_PURPLE);

  canvas->setCursor(x, y);
  canvas->setTextColor(ILI9341_YELLOW);
  canvas->setTextSize(3);
  canvas->print("MCH2022");

  canvas->setCursor(0, 0);
  canvas->setTextColor(ILI9341_WHITE);
  canvas->setTextSize(1);
  canvas->print(voltage);
  canvas->print(" (V)");
  if (charging) {
    canvas->print(" charging");
  }

  /* Move the text for the next round */
  x += 16;
  if (x >= ILI9341_TFTHEIGHT) {
    y += 24;
    x = 0;
    if (y >= ILI9341_TFTWIDTH) {
      y = 0;
    }
  }

  tft.drawRGBBitmap(0, 0, canvas->getBuffer(), canvas->width(), canvas->height());
}

void init_leds_after_display() {
  FastLED.addLeds<SK6812, GPIO_LED_DATA, GRB>(leds, NUM_LEDS).setCorrection(TypicalLEDStrip);
  FastLED.setBrightness(96);

  /* This has to be placed after SPI (LCD) has been initialized. The
   * Arduino library wants to use this pin as SPI MISO... */
  pinMode(GPIO_LED_ENABLE, OUTPUT);
  digitalWrite(GPIO_LED_ENABLE, HIGH);
}

/**
 * Update the kite LEDs. Loop over the colors of the rainbow.
 */
void change_leds() {
  /* This would set the first LED to white. */
  //leds[0] = CRGB::White;

  /* This merely sets all of the LEDs local storage (leds[]) to the new value. */
  for (int i = 0; i < NUM_LEDS; ++i) {
    led_color[led_reorder[i]].hue = ((led_color[led_reorder[i]].hue + 1) % 256);
    hsv2rgb_rainbow(led_color[i], leds[i]);
  }

  /* This uses the leds[] values and notifies the SK6812 of the new colors. */
  FastLED.show();
}

/**
 * Get charging state.
 */
void get_battery_state(bool& charging, float& voltage) {
  const float conversion_factor = 3.3f / (1 << 12); /* 12-bit ADC with 3.3v vref */
  uint16_t raw_voltage;
  Wire.beginTransmission(RP2040_ADDR);
  Wire.write(RP2040_REG_ADC_VALUE_VBAT_LO);
  Wire.endTransmission();
  Wire.requestFrom(RP2040_ADDR, 2);
  raw_voltage = (Wire.read() | (Wire.read() << 8));
  voltage = raw_voltage * conversion_factor * 2; /* Connected through 100k/100k divider */

  Wire.beginTransmission(RP2040_ADDR);
  Wire.write(RP2040_REG_CHARGING_STATE);
  Wire.endTransmission();
  Wire.requestFrom(RP2040_ADDR, 2);
  charging = (bool)Wire.read();
  raw_voltage = (Wire.read() | (Wire.read() << 8));
}

/**
 * Backlight brightness can be set from 0 (off) to 255.
 */
void set_backlight(uint8_t brightness) {
    Wire.beginTransmission(RP2040_ADDR);
    Wire.write(RP2040_REG_LCD_BACKLIGHT);
    Wire.write(brightness);
    Wire.endTransmission();
}

/**
 * Supports NEC remote control codes. See:
 * https://en.wikipedia.org/wiki/Consumer_IR
 * https://github.com/flipperdevices/flipperzero-firmware/blob/
 *   52b59662627254ae5ccde6201b5c4921573fc769/documentation/file_formats/InfraredFileFormats.md
 *
 * Send simple NEC address as 0x0004 or extended one with high byte second 0xHHLL.
 *
 * Right now, the RP2040 only does (extended) NEC commands. No (Philips) RC-5.
 */
void send_ir_nec(uint16_t address, uint8_t command) {
  Wire.beginTransmission(RP2040_ADDR);
  Wire.write(RP2040_REG_IR_ADDRESS_LO);
  Wire.write(address & 0xFF);
  if ((address >> 8) == 0) {
    Wire.write(~address);
  } else {
    Wire.write(address >> 8);
  }
  Wire.write(command);
  Wire.write(0x01); /* trigger */
  Wire.endTransmission();
}

/**
 * Exit application and return to launched by resetting the ESP.
 */
void return_to_launcher() {
  REG_WRITE(RTC_CNTL_STORE0_REG, 0);
  esp_restart();
}
